#region License Information
/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;
using HEAL.Attic;
using DiffSharp.Interop.Float64;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  [StorableType("7BC8666F-DB02-405A-9567-8378C7BBDE9A")]
  [Item("NonlinearLeastSquaresVectorConstantOptimizationEvaluator", "")]
  public class NonlinearLeastSquaresVectorConstantOptimizationEvaluator : SymbolicRegressionConstantOptimizationEvaluator {

    private const string ConstantOptimizationIterationsName = "ConstantOptimizationIterations";

    #region Parameter Properties
    public IFixedValueParameter<IntValue> ConstantOptimizationIterationsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[ConstantOptimizationIterationsName]; }
    }
    #endregion

    #region Properties
    public int ConstantOptimizationIterations {
      get { return ConstantOptimizationIterationsParameter.Value.Value; }
      set { ConstantOptimizationIterationsParameter.Value.Value = value; }
    }
    #endregion

    public NonlinearLeastSquaresVectorConstantOptimizationEvaluator()
      : base() {
      Parameters.Add(new FixedValueParameter<IntValue>(ConstantOptimizationIterationsName, "Determines how many iterations should be calculated while optimizing the constant of a symbolic expression tree(0 indicates other or default stopping criterion).", new IntValue(10)));
    }

    protected NonlinearLeastSquaresVectorConstantOptimizationEvaluator(NonlinearLeastSquaresVectorConstantOptimizationEvaluator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new NonlinearLeastSquaresVectorConstantOptimizationEvaluator(this, cloner);
    }
    [StorableConstructor]
    protected NonlinearLeastSquaresVectorConstantOptimizationEvaluator(StorableConstructorFlag _) : base(_) { }

    protected override ISymbolicExpressionTree OptimizeConstants(
      ISymbolicExpressionTree tree, IRegressionProblemData problemData, IEnumerable<int> rows,
      CancellationToken cancellationToken = default(CancellationToken), EvaluationsCounter counter = null) {
      return OptimizeTree(tree,
        problemData, rows,
        ApplyLinearScalingParameter.ActualValue.Value, ConstantOptimizationIterations, UpdateVariableWeights,
        cancellationToken, counter);
    }

    public static ISymbolicExpressionTree OptimizeTree(
      ISymbolicExpressionTree tree,
      IRegressionProblemData problemData, IEnumerable<int> rows,
      bool applyLinearScaling, int maxIterations, bool updateVariableWeights,
      CancellationToken cancellationToken = default(CancellationToken), EvaluationsCounter counter = null, Action<double[], double, object> iterationCallback = null) {



      DV targets = new DV(problemData.Dataset.GetDoubleValues(problemData.TargetVariable, rows).ToArray());
      DV Predictions(DV variables) {
        var predictions = rows.Select(row => {
          var scalarParams = problemData.AllowedInputVariables.Intersect(problemData.Dataset.DoubleVariables)
            .ToDictionary(var => var, var => new D(problemData.Dataset.GetDoubleValue(var, row)));
          var vectorParams = problemData.AllowedInputVariables.Intersect(problemData.Dataset.DoubleVectorVariables)
            .ToDictionary(var => var, var => new DV(problemData.Dataset.GetDoubleVectorValue(var, row).ToArray()));

          return TreeToDiffSharpConverter.Evaluate(tree,
            updateVariableWeights, applyLinearScaling,
            variables,
            scalarParams, vectorParams);
        }).ToArray();
        return new DV(predictions);
      }
      D Mse(DV vars) => DV.Mean(DV.Pow(targets - Predictions(vars), 2));

      Func<DV, D> loss = Mse;
      Func<DV, DV> lossGrad = AD.Grad(loss);
      Func<DV, DM> lossHess = AD.Hessian(loss);

      var initialConstants = TreeToDiffSharpConverter.GetInitialConstants(tree,
        updateVariableWeights, applyLinearScaling,
        problemData.AllowedInputVariables.Intersect(problemData.Dataset.DoubleVariables)
          .ToDictionary(var => var, var => new D(problemData.Dataset.GetDoubleValue(var, rows.First()))),
        problemData.AllowedInputVariables.Intersect(problemData.Dataset.DoubleVectorVariables)
          .ToDictionary(var => var, var => new DV(problemData.Dataset.GetDoubleVectorValue(var, rows.First()).ToArray())));


      //extract initial constants
      double[] c = initialConstants.ToArray();

      alglib.ndimensional_func alglib_func = (double[] x, ref double func, object o) => {
        var vars = new DV(x);
        func = (double)loss(vars);
        var cnt = (EvaluationsCounter)o;
        cnt.FunctionEvaluations++;
      };
      alglib.ndimensional_grad alglib_grad = (double[] x, ref double func, double[] grad, object o) => {
        var vars = new DV(x);
        func = (double)loss(vars);
        Array.Copy((double[])lossGrad(vars), grad, x.Length);
        var cnt = (EvaluationsCounter)o;
        cnt.GradientEvaluations++;
      };
      alglib.ndimensional_hess alglib_hess = (double[] x, ref double func, double[] grad, double[,] hess, object o) => {
        var vars = new DV(x);
        func = (double)loss(vars);
        Array.Copy((double[])lossGrad(vars), grad, x.Length);
        Array.Copy((double[,])lossHess(vars), hess, x.Length * x.Length);
        var cnt = (EvaluationsCounter)o;
        cnt.HessianEvaluations++;
      };
      alglib.ndimensional_rep xrep = (p, f, obj) => {
        iterationCallback?.Invoke(p, f, obj);
        cancellationToken.ThrowIfCancellationRequested();
      };
      var rowEvaluationsCounter = new EvaluationsCounter();

      try {

        alglib.minlmcreatefgh(c, out var state);
        alglib.minlmsetcond(state, 0.0, 0.0, 0.0, maxIterations);
        alglib.minlmsetxrep(state, iterationCallback != null || cancellationToken != default(CancellationToken));
        alglib.minlmoptimize(state, alglib_func, alglib_grad, alglib_hess, xrep, rowEvaluationsCounter);
        alglib.minlmresults(state, out c, out var rep);

        ////retVal == -7  => constant optimization failed due to wrong gradient
        //if (retVal == -1)
        //  return (ISymbolicExpressionTree)tree.Clone();
      } catch (ArithmeticException) {
        return (ISymbolicExpressionTree)tree.Clone();
      } catch (alglib.alglibexception) {
        return (ISymbolicExpressionTree)tree.Clone();
      }

      if (counter != null) {
        int n = rows.Count();
        counter.FunctionEvaluations += rowEvaluationsCounter.FunctionEvaluations / n;
        counter.GradientEvaluations += rowEvaluationsCounter.GradientEvaluations / n;
        counter.HessianEvaluations += rowEvaluationsCounter.HessianEvaluations / n;
      }

      var newTree = (ISymbolicExpressionTree)tree.Clone();
      UpdateConstants(newTree, c, updateVariableWeights);

      return newTree;
    }

    private static void UpdateConstants(ISymbolicExpressionTree tree, double[] constants, bool updateVariableWeights) {
      int i = 0;
      foreach (var node in tree.Root.IterateNodesPrefix().OfType<SymbolicExpressionTreeTerminalNode>()) {
        ConstantTreeNode constantTreeNode = node as ConstantTreeNode;
        VariableTreeNodeBase variableTreeNodeBase = node as VariableTreeNodeBase;
        FactorVariableTreeNode factorVarTreeNode = node as FactorVariableTreeNode;
        if (constantTreeNode != null)
          constantTreeNode.Value = constants[i++];
        else if (updateVariableWeights && variableTreeNodeBase != null)
          variableTreeNodeBase.Weight = constants[i++];
        else if (factorVarTreeNode != null) {
          for (int j = 0; j < factorVarTreeNode.Weights.Length; j++)
            factorVarTreeNode.Weights[j] = constants[i++];
        }
      }
    }

    public static bool CanOptimizeConstants(ISymbolicExpressionTree tree) {
      return TreeToDiffSharpConverter.IsCompatible(tree);
    }
  }
}