#region License Information
/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;
using HEAL.Attic;
using DoubleVector = MathNet.Numerics.LinearAlgebra.Vector<double>;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  [StorableType("5F8DB251-C6F7-40AC-BC30-3C55AF655A2F")]
  [Item("VectorUnrollingNonlinearLeastSquaresConstantOptimizationEvaluator", "")]
  public class VectorUnrollingNonlinearLeastSquaresConstantOptimizationEvaluator : SymbolicRegressionConstantOptimizationEvaluator {

    private const string ConstantOptimizationIterationsName = "ConstantOptimizationIterations";

    #region Parameter Properties
    public IFixedValueParameter<IntValue> ConstantOptimizationIterationsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[ConstantOptimizationIterationsName]; }
    }
    #endregion

    #region Properties
    public int ConstantOptimizationIterations {
      get { return ConstantOptimizationIterationsParameter.Value.Value; }
      set { ConstantOptimizationIterationsParameter.Value.Value = value; }
    }
    #endregion

    public VectorUnrollingNonlinearLeastSquaresConstantOptimizationEvaluator()
      : base() {
      Parameters.Add(new FixedValueParameter<IntValue>(ConstantOptimizationIterationsName, "Determines how many iterations should be calculated while optimizing the constant of a symbolic expression tree(0 indicates other or default stopping criterion).", new IntValue(10)));
    }

    protected VectorUnrollingNonlinearLeastSquaresConstantOptimizationEvaluator(VectorUnrollingNonlinearLeastSquaresConstantOptimizationEvaluator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new VectorUnrollingNonlinearLeastSquaresConstantOptimizationEvaluator(this, cloner);
    }
    [StorableConstructor]
    protected VectorUnrollingNonlinearLeastSquaresConstantOptimizationEvaluator(StorableConstructorFlag _) : base(_) { }

    protected override ISymbolicExpressionTree OptimizeConstants(
      ISymbolicExpressionTree tree, IRegressionProblemData problemData, IEnumerable<int> rows,
      CancellationToken cancellationToken = default(CancellationToken), EvaluationsCounter counter = null) {
      return OptimizeTree(tree, (SymbolicDataAnalysisExpressionTreeVectorInterpreter)SymbolicDataAnalysisTreeInterpreterParameter.ActualValue,
        problemData, rows,
        ApplyLinearScalingParameter.ActualValue.Value, ConstantOptimizationIterations, UpdateVariableWeights,
        cancellationToken, counter);
    }

    public static ISymbolicExpressionTree OptimizeTree(
      ISymbolicExpressionTree tree,
      SymbolicDataAnalysisExpressionTreeVectorInterpreter interpreter,
      IRegressionProblemData problemData, IEnumerable<int> rows,
      bool applyLinearScaling, int maxIterations, bool updateVariableWeights,
      CancellationToken cancellationToken = default(CancellationToken), EvaluationsCounter counter = null, Action<double[], double, object> iterationCallback = null) {

      var vectorLengths = problemData.Dataset.DoubleVectorVariables
        .SelectMany(var => problemData.Dataset.GetDoubleVectorValues(var, rows))
        .Select(v => v.Count);
      var vectorlength = vectorLengths.First();
      if (vectorLengths.Any(l => l != vectorlength))
        throw new InvalidOperationException("All vectors must be of same length.");
      var evaluationTraces = interpreter.GetIntermediateNodeValues(tree, problemData.Dataset, rows);
      var evaluationTrace = evaluationTraces.First(); // assume all vector lengths are the same


      // numeric constants in the tree become variables for constant opt
      // variables in the tree become parameters (fixed values) for constant opt
      // for each parameter (variable in the original tree) we store the 
      // variable name, variable value (for factor vars) and lag as a DataForVariable object.
      // A dictionary is used to find parameters
      bool success = VectorUnrollingTreeToAutoDiffTermConverter.TryConvertToAutoDiff(
        tree, evaluationTrace,
        updateVariableWeights, applyLinearScaling,
        out var parameters, out var initialConstants, out var func, out var func_grad);
      if (!success)
        throw new NotSupportedException("Could not optimize constants of symbolic expression tree due to not supported symbols used in the tree.");
      if (parameters.Count == 0) return (ISymbolicExpressionTree)tree.Clone();
      var parameterEntries = parameters.ToArray(); // order of entries must be the same for x

      //extract initial constants
      double[] c;
      if (applyLinearScaling) {
        c = new double[initialConstants.Length + 2];
        c[0] = 0.0;
        c[1] = 1.0;
        Array.Copy(initialConstants, 0, c, 2, initialConstants.Length);
      } else {
        c = (double[])initialConstants.Clone();
      }

      IDataset ds = problemData.Dataset;
      double[,] x = new double[rows.Count(), parameters.Count];
      int row = 0;
      foreach (var r in rows) {
        int col = 0;
        foreach (var info in parameterEntries) {
          if (ds.VariableHasType<double>(info.variableName)) {
            x[row, col] = ds.GetDoubleValue(info.variableName, r + info.lag);
          } else if (ds.VariableHasType<string>(info.variableName)) {
            x[row, col] = ds.GetStringValue(info.variableName, r) == info.variableValue ? 1 : 0;
          } else if (ds.VariableHasType<DoubleVector>(info.variableName)) {
            x[row, col] = ds.GetDoubleVectorValue(info.variableName, r)[info.index];
          } else throw new InvalidProgramException("found a variable of unknown type");
          col++;
        }
        row++;
      }
      double[] y = ds.GetDoubleValues(problemData.TargetVariable, rows).ToArray();
      int n = x.GetLength(0);
      int m = x.GetLength(1);
      int k = c.Length;

      alglib.ndimensional_pfunc function_cx_1_func = CreatePFunc(func);
      alglib.ndimensional_pgrad function_cx_1_grad = CreatePGrad(func_grad);
      alglib.ndimensional_rep xrep = (p, f, obj) => {
        iterationCallback?.Invoke(p, f, obj);
        cancellationToken.ThrowIfCancellationRequested();
      };
      var rowEvaluationsCounter = new EvaluationsCounter();

      try {
        alglib.lsfitcreatefg(x, y, c, n, m, k, false, out var state);
        alglib.lsfitsetcond(state, 0.0, 0.0, maxIterations);
        alglib.lsfitsetxrep(state, iterationCallback != null || cancellationToken != default(CancellationToken));
        //alglib.lsfitsetgradientcheck(state, 0.001);
        alglib.lsfitfit(state, function_cx_1_func, function_cx_1_grad, xrep, rowEvaluationsCounter);
        alglib.lsfitresults(state, out var retVal, out c, out alglib.lsfitreport rep);

        //retVal == -7  => constant optimization failed due to wrong gradient
        if (retVal == -1)
          return (ISymbolicExpressionTree)tree.Clone();
      } catch (ArithmeticException) {
        return (ISymbolicExpressionTree)tree.Clone();
      } catch (alglib.alglibexception) {
        return (ISymbolicExpressionTree)tree.Clone();
      }

      if (counter != null) {
        counter.FunctionEvaluations += rowEvaluationsCounter.FunctionEvaluations / n;
        counter.GradientEvaluations += rowEvaluationsCounter.GradientEvaluations / n;
      }

      var newTree = (ISymbolicExpressionTree)tree.Clone();
      if (applyLinearScaling) {
        var tmp = new double[c.Length - 2];
        Array.Copy(c, 2, tmp, 0, tmp.Length);
        UpdateConstants(newTree, tmp, updateVariableWeights);
      } else
        UpdateConstants(newTree, c, updateVariableWeights);

      return newTree;
    }

    private static void UpdateConstants(ISymbolicExpressionTree tree, double[] constants, bool updateVariableWeights) {
      int i = 0;
      foreach (var node in tree.Root.IterateNodesPrefix().OfType<SymbolicExpressionTreeTerminalNode>()) {
        ConstantTreeNode constantTreeNode = node as ConstantTreeNode;
        VariableTreeNodeBase variableTreeNodeBase = node as VariableTreeNodeBase;
        FactorVariableTreeNode factorVarTreeNode = node as FactorVariableTreeNode;
        if (constantTreeNode != null)
          constantTreeNode.Value = constants[i++];
        else if (updateVariableWeights && variableTreeNodeBase != null)
          variableTreeNodeBase.Weight = constants[i++];
        else if (factorVarTreeNode != null) {
          for (int j = 0; j < factorVarTreeNode.Weights.Length; j++)
            factorVarTreeNode.Weights[j] = constants[i++];
        }
      }
    }

    private static alglib.ndimensional_pfunc CreatePFunc(VectorUnrollingTreeToAutoDiffTermConverter.ParametricFunction func) {
      return (double[] c, double[] x, ref double fx, object o) => {
        fx = func(c, x);
        var counter = (EvaluationsCounter)o;
        counter.FunctionEvaluations++;
      };
    }

    private static alglib.ndimensional_pgrad CreatePGrad(VectorUnrollingTreeToAutoDiffTermConverter.ParametricFunctionGradient func_grad) {
      return (double[] c, double[] x, ref double fx, double[] grad, object o) => {
        var tuple = func_grad(c, x);
        fx = tuple.Item2;
        Array.Copy(tuple.Item1, grad, grad.Length);
        var counter = (EvaluationsCounter)o;
        counter.GradientEvaluations++;
      };
    }

    public static bool CanOptimizeConstants(ISymbolicExpressionTree tree) {
      return VectorUnrollingTreeToAutoDiffTermConverter.IsCompatible(tree);
    }
  }
}